#include <stdint.h>
#include <stddef.h>
#include "../pdl.h"

typedef int32_t fixed_t;
#define NB_POINT_CUBE 8
#define NB_LINES_CUBE 12

#define PRECISION 8
#define TO_FIXED(num) (int)((num)*(1<<PRECISION))
#define TO_INT(num) ((num)>>PRECISION)
#define MUL(a, b) (((a)*(b))>>PRECISION)
#define DIV(a, b) (((a)<<PRECISION)/(b))
#define SGN(a) ((a) < 0 ? -1 : (a) > 0 ? 1 : 0)

uint8_t Screen256[ 88704 ];

fixed_t angle = 0;

typedef struct
{
  fixed_t x;
  fixed_t y;
  fixed_t z;
  int16_t sx;
  int16_t sy;
} Vertex3D;

typedef struct
{
    int8_t p1;
    int8_t p2;
} Pairs;

/* Line info, expressed as byte offset from the pointer [Cube] to the [sx/sy]
   pair of each line. Basically instead of saying line i we go for the address
   of Cube[i].sx relative to Cube, which makes the drawing code shorter. */
#define OFFSET(N) (sizeof(Vertex3D) * (N) + offsetof(Vertex3D, sx))
Pairs Lines[NB_LINES_CUBE] = {
    { OFFSET(0), OFFSET(1) },
    { OFFSET(1), OFFSET(2) },
    { OFFSET(2), OFFSET(3) },
    { OFFSET(3), OFFSET(0) },

    { OFFSET(4), OFFSET(5) },
    { OFFSET(5), OFFSET(6) },
    { OFFSET(6), OFFSET(7) },
    { OFFSET(7), OFFSET(4) },

    { OFFSET(0), OFFSET(4) },
    { OFFSET(1), OFFSET(5) },
    { OFFSET(2), OFFSET(6) },
    { OFFSET(3), OFFSET(7) },
};
#undef OFFSET
Vertex3D Cube[NB_POINT_CUBE];


fixed_t sine(fixed_t d) {
    /* Argument reduction: we need to mod d %= 360 and again mod d %= 180, and
       remember whether the last made a difference. We do this by repeatedly
       subtracting 180 and counting whether the number of subtractions is odd
       by making "swap" alternate between 1 and -1. We also do this in a do/
       while loop, so we end up with one extra iteration so at the end d < 0.
       This is useful because we need that to compute 180-x anyway. */
    int swap = -1;
    int x = TO_FIXED(180);
    do d -= x, swap = -swap;
    while(d >= 0);

    fixed_t para = MUL(d + x, -d);
    fixed_t v = DIV(4 * para, (TO_FIXED(40500) - para));
    /* Should be v * swap (so v or -v). However, v ^ swap gives us v^1 and
       v^-1, which is easier to compute and in both cases differs exactly by
       the LSB. This is an acceptable approximation. */
    return v ^ swap;
}

fixed_t cosine(fixed_t d) {
    return sine( d + TO_FIXED(90) );
}

void RotateAndProjectVertex( Vertex3D *vertex )
{
    // Ruse du cowboy car dans notre cas ax=0 et ay=az=angle;
	fixed_t ry = vertex->y;
	fixed_t rz = vertex->z;
    fixed_t C = cosine(angle), S = sine(angle);
	fixed_t rx = MUL(vertex->x, C) + MUL(rz, S);
	rz =  MUL(rz, C) - MUL(vertex->x, S);
    fixed_t tmpx = MUL(rx, C) - MUL(ry, S);
	ry = MUL(rx, S) + MUL(ry, C);
	rx = tmpx;

    fixed_t divisor = DIV( TO_FIXED(250), (rz +  TO_FIXED(4.16667)) );
    vertex->sx = TO_INT( MUL( rx, divisor ) );
    vertex->sy = TO_INT( MUL( ry, divisor ) );
}

void Drawline( uint8_t *buffer, int x1, int y1, int x2, int y2, uint8_t color)
{
	int i;
	int dx = x2 - x1;
    int dy = y2 - y1;
	int sx = SGN(dx);
    int sy = SGN(dy);

    buffer += 396 * (y1 + 112) + (x1 + 198);

	dx = (dx >= 0 ? dx : -dx);
    dy = (dy >= 0 ? dy : -dy);

    if(dx < dy) {
        int tmp = dx;
        dx = dy;
        dy = tmp;
        tmp = sx;
        sx = 396 * sy;
        sy = tmp;
    }
    else {
        sy *= 396;
    }

	int cumul = dx >> 1;

	for(i = 0; i <= dx; i++)
	{
		buffer += sx;
		cumul += dy;
		if(cumul > dx) cumul -= dx, buffer += sy;
        *buffer = color;
	}
}

void Blur( uint8_t *buffer )
{
    buffer += 88308;
    
    for( int index = 0; index < 223; index++ )
    {
        for( int x = 0;  x < 394; x++ )
        {
            int temp = buffer[0] +
                       buffer[1] +
                       buffer[2] +
                       buffer[-395];
            buffer[-395] = temp >> 2;
            buffer++;
        }
        buffer -= 395 + 395;
    }
}

void Dclear( uint8_t *Screen, uint8_t color )
{
	for(int i = 0; i < 88704; i++)
		Screen[i] = color;
}

void Render( uint16_t *vram, uint8_t *Screen )
{
	for(int i = 0; i < 88704; i++) {
        int u = Screen[i] & 63;
        int x = 0xffe0 + (u >> 1);
        const int8_t shifts[] = { 16, 11, 6, 0 };
		vram[i] = x << shifts[Screen[i] >> 6];
    }
}

void FirecubeInit( void )
{
    /* Initialize longwords of Cube (4 per vertex) to either -0.75 or 0.75,
       as indicated by one bit of initMask. We initialize the pair sx/sy using
       a useless bit. sx/sy is normally uninitialized, so we don't care. We
       just set the last pair to 1 so the do/while loop doesn't stop early.

         Order:    x  y  z  sx/sy
         Cube[0]:  -  -  -  0
         Cube[1]:  +  -  -  0
         Cube[2]:  +  +  -  0
         Cube[3]:  -  +  -  0
         Cube[4]:  -  -  +  0
         Cube[5]:  +  -  +  0
         Cube[6]:  +  +  +  0
         Cube[7]:  -  +  +  1

       The order is given from MSB to LSB as our working bit is the MSB.*/
    int initMask = 0b11100110001010101100010000001001;
    int *ptr = (int *)Cube;

    do {
        /* We want -0.75 if initMask is negative. This differs only by the LSB,
           which is a slight but invisible approximation. */
        *ptr++ = TO_FIXED(0.75) ^ (initMask >> 31);
        initMask <<= 1;
    }
    while(initMask);
}

void FirecubeUpdate( void )
{
    /* No reduction modulo 360 needed. sine() does it when called, and the
       animation doesn't run long enough to impact performance or risk an
       overflow. */
	angle += TO_FIXED(1);
    for (int i = 0; i < NB_POINT_CUBE; i++)
		RotateAndProjectVertex(&Cube[i]);
}

void FirecubeRender( uint8_t *screen )
{ 
    for (int j = 0; j < NB_LINES_CUBE; j++) {
        int16_t *s1 = (void *)Cube + Lines[j].p1;
        int16_t *s2 = (void *)Cube + Lines[j].p2;
        Drawline(screen, s1[0], s1[1], s2[0], s2[1], 255);
    }
    Blur( screen );
}

PDL_MAIN int main(void)
{
	Dclear( Screen256, 0 );

	FirecubeInit( );

    do {
		FirecubeUpdate( );
		FirecubeRender( Screen256 );
		Render( PDL->vram, Screen256 );
		PDL->dupdate();
	}
	while(PDL->time_ms <= 10000);

	return 0;
}
